(*#######################################################################################

UpdateLoader - Protokollinformationen und Standardeinstellungen

http://www.leo-andres.de/

Copyright (C) 2012 Leo-Andres Hofmann


This file is part of UpdateLoader.

UpdateLoader is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

UpdateLoader is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with UpdateLoader. If not, see <http://www.gnu.org/licenses/>.

#######################################################################################*)
unit protocol;

{$mode objfpc}{$H+}

interface

uses
  Classes, SysUtils, FileUtil, Forms, Controls, Graphics, Dialogs, ComCtrls,
  ExtCtrls, EditBtn, Buttons, LCLType, LCLIntf, Tools, SynaSer;

const
  //Standardeinstellungen
  defaultConnectTries: integer = 250;
  defaultBootloaderPassword: string = 'Peda';
  defaultDeviceName: string = 'Unbekannt';
  defaultBaudrate: integer = 7; //Eintrags-Nr. in der ComboBox

  //Einstellungen für die Übertragung
  settingValidBootloader: string = '2.1';
  settingConnectTryTimeout: integer = 25; //Wartezeit auf Antwort beim Verbinden
  settingPortSetupDelay: integer = 500; //Wartezeit beim Öffnen des COM-Ports
  settingResetToggleDelay: integer = 50; //Wartezeit zur Reset-Auslösung über DTR
  settingConnectMessageDelay: integer = 150; //Wartezeit zwischen GUI-Anzeigen
  settingReadTimeout: integer = 150; //Allgemeine Wartezeit auf serielle Daten
  settingReadLongTimeout: integer = 500; //Wartezeit auf "continue" beim Flashen
  settingMaximumBufferSize: longint = 262144;
  settingVerifyChunkLength: integer = 512;
  settingMinimumWriteBuffer: integer = 128;

type
  TBootloaderRoutines = class
    constructor Create;
    destructor Destroy; override;
    private
      const
         deviceNames: array[0..22] of string = (
           '1E 90 07:ATtiny13',
           '1E 91 0A:ATtiny2313',
           '1E 92 05:ATmega48',
           '1E 92 06:ATtiny45',
           '1E 92 07:ATtiny44',
           '1E 92 08:ATtiny461',
           '1E 93 06:ATmega8515',
           '1E 93 07:ATmega8',
           '1E 93 08:ATmega8535',
           '1E 93 0A:ATmega88',
           '1E 93 0B:ATtiny85',
           '1E 93 0C:ATtiny84',
           '1E 93 0D:ATtiny861',
           '1E 93 0F:ATmega88P',
           '1E 94 03:ATmega16',
           '1E 94 04:ATmega162',
           '1E 94 06:ATmega168',
           '1E 94 0B:ATmega168P',
           '1E 95 01:ATmega323',
           '1E 95 02:ATmega32',
           '1E 95 0F:ATmega328P',
           '1E 96 09:ATmega644',
           '1E 98 02:ATmega2561');

         //Bootloader-Kommunikation
         blcIdentifierCommand: byte = $A5;
         blcIdentifierEscape: byte = $A5;
         blcIdentifierAnswer: byte = $A8;
         blcIdentifierContinue: byte = $A9;

         blcReadRevision: byte = $0;
         blcReadBufferSize: byte = $1;
         blcReadSignature: byte = $2;
         blcReadFlashSize: byte = $3;

         blcCommandProgramWrite: byte = $4;
         blcCommandProgramStart: byte = $5;
         blcCommandProgramCheckCRC: byte = $6;
         blcCommandProgramVerify: byte = $7;
         blcCommandProgramFinish: byte = $80;

         blcStatusConnected: byte = $A6;
         blcStatusBadCommand: byte = $A7;
         blcStatusSuccess: byte = $AA;
         blcStatusFail: byte = $AB;

         blcDataIllegalChar: byte = $13;
         blcDataIllegalCharShift: byte = $80;
         blcDataPasswordLeader: byte = $0D;
         blcDataPasswordTrailer: byte = $FF;

      var
         serialConnection: TBlockSerial;
         connectionCRC: word;
         oneWire: boolean;
         failureAddress: longint; //Zuletzt bearbeiter Speicherbereich

      procedure serialSendByte(data: byte);
      procedure serialSendString(data: string);

      procedure calculateCRC(data: byte);
      procedure skipOneWireBytes(count: integer);

      function detectSupport(command: byte): boolean;
      function readInfo(readCommand: integer): TByteList;
      function uploadData(targetCommand: byte; waitForContinue: boolean; buffer: TMemoryBuffer; bufferLength: longint; chunkLength: longint; progressBar: TProgressBar): boolean;

      function getBufferChunk(buffer: TMemoryBuffer; var startPosition: longint; chunkLength: longint; bufferMaxLength: longint): string;

    public
      procedure assignSerialConnection(connection: TBlockSerial);
      procedure setWireMode(mode: boolean);
      procedure resetCRC;

      function connect(bootloaderPassword: string; connectTries: integer; toggleDTR: boolean = true; detectOneWire: boolean = true; progressBar: TProgressBar = nil): integer;
      function writeFirmware(buffer: TMemoryBuffer; bufferLength: longint; chunkLength: longint; progressBar: TProgressBar = nil): boolean;
      function verifyFirmware(buffer: TMemoryBuffer; bufferLength: longint; progressBar: TProgressBar = nil): boolean;
      function startFirmware: boolean;

      function checkCRC: boolean;

      function detectVerifySupport: boolean;
      function detectCRCSupport: boolean;

      function readRevision: string;
      function readSignature: string;
      function readBufferSize: integer;
      function readFlashSize: longint;

      function getFailureAddress: string;
      function getDeviceName(signature: string): string;
  end;

implementation

{$REGION 'Initialisierung'}
constructor TBootloaderRoutines.Create;
begin
  inherited;

  connectionCRC := 0;
  oneWire := false;
  failureAddress := 0;
end;

destructor TBootloaderRoutines.Destroy;
begin
  inherited;
end;

procedure TBootloaderRoutines.assignSerialConnection(connection: TBlockSerial);
begin
  self.serialConnection := connection;
end;

procedure TBootloaderRoutines.setWireMode(mode: boolean);
begin
  self.oneWire := mode;
end;

procedure TBootloaderRoutines.resetCRC;
begin
  self.connectionCRC := 0;
end;
{$ENDREGION}

{$REGION 'Hauptfunktionen (Übertragen, Verbinden)'}
function TBootloaderRoutines.connect(bootloaderPassword: string; connectTries: integer; toggleDTR: boolean = true; detectOneWire: boolean = true; progressBar: TProgressBar = nil): integer;
var
  connectTry: integer;
  wireMode: byte;
  wireDetectChar: byte;
  reply: byte;
  connectionEstablished: boolean;
begin
  wireMode := 0;
  wireDetectChar := ord(bootloaderPassword[1]);
  connectionEstablished := false;
  Result := 0;

  //Auf gültiges Passwort prüfen
  if pos(AnsiChar(blcStatusConnected), bootloaderPassword) <> 0 then
  begin
    raise Exception.Create('Ungültiges Passwort, es dürfen keine Protokollzeichen im Passwort vorkommen');
    exit;
  end;

  if Assigned(progressBar) then
  begin
    progressBar.Position := 0;
    progressBar.Max := connectTries;
    progressBar.Step := 1;
  end;

  //Reset-Pin Ansteuerung über DTR
  //Fallende Flanke am DTR-Pin erzeugen (Reset auslösen), Pin danach auf High-Pegel halten
  //Pin gesetzt (true) = Low-Pegel!
  if toggleDTR then
  begin
    serialConnection.DTR := false;
    Sleep(settingResetToggleDelay);
    serialConnection.DTR := true;
    Sleep(settingResetToggleDelay);
    serialConnection.DTR := false;
  end;

  //Puffer leeren
  serialConnection.Purge;
  serialConnection.Flush;

  for connectTry := 0 to connectTries do
  begin

    //0x0D + Passwort + 0xFF senden
    serialSendByte(blcDataPasswordLeader);
    serialSendString(bootloaderPassword);
    serialSendByte(blcDataPasswordTrailer);

    try
      if serialConnection.CanRead(settingConnectTryTimeout) = true then
      begin
        repeat
          reply := serialConnection.RecvByte(settingConnectTryTimeout);

          if (reply = wireDetectChar) and (detectOneWire) then
             wireMode := 1;

          if reply = blcStatusConnected then
          begin
            connectionEstablished := true;
            if (wireMode <> 1) or (not detectOneWire) then //keine 1-Draht Verbindung oder Erkennung deaktiviert -> 2-Draht
               wireMode := 2;

            break;
          end;
        until serialConnection.WaitingData <= 0;
      end;
    except
    end;

    if connectionEstablished = true then
       break;

    if Assigned(progressBar) then
      progressBar.StepIt;
    Application.ProcessMessages;
  end;

  serialConnection.Purge;
  serialConnection.Flush;

  if connectionEstablished = true then
  begin
    serialSendByte(blcIdentifierCommand);

    try
      //Eigene Daten im 1-Wire Modus verwerfen
      if wireMode = 1 then
        serialConnection.RecvByte(settingReadTimeout);

      reply := serialConnection.RecvByte(settingReadTimeout);
    except
    end;

    serialConnection.Purge;
    serialConnection.Flush;

    //Antwort nur prüfen wenn eine Antwort empfangen wurde
    if (reply <> blcStatusSuccess) and (serialConnection.LastError <= 0) then
    begin
      raise Exception.Create('Ungültige Antwort des Bootloaders (0x' + IntToHex(reply, 2) + ')');
      exit;
    end;
  end
  else
  begin
    raise Exception.Create('Zeitüberschreitung');
    exit;
  end;

  Result := wireMode;

  if Assigned(progressBar) then
    progressBar.Position := progressBar.Max;
  Application.ProcessMessages;
end;

function TBootloaderRoutines.writeFirmware(buffer: TMemoryBuffer; bufferLength: longint; chunkLength: longint; progressBar: TProgressBar = nil): boolean;
begin
  Result := uploadData(blcCommandProgramWrite, true, buffer, bufferLength, chunkLength, progressBar);
end;

function TBootloaderRoutines.verifyFirmware(buffer: TMemoryBuffer; bufferLength: longint; progressBar: TProgressBar = nil): boolean;
begin
  Result := uploadData(blcCommandProgramVerify, false, buffer, bufferLength, settingVerifyChunkLength, progressBar);
end;

function TBootloaderRoutines.startFirmware: boolean;
begin
  Result := true;

  try
    serialSendByte(blcIdentifierCommand);
    serialSendByte(blcCommandProgramStart);

    skipOneWireBytes(2);

    try
      if serialConnection.RecvByte(settingReadTimeout) = blcStatusBadCommand then
        Result := false;
    except
    end;

  finally
    serialConnection.Purge;
    serialConnection.Flush;
  end;
end;
{$ENDREGION}

{$REGION 'CRC-Prüfsumme'}
function TBootloaderRoutines.checkCRC: boolean;
var
  reply: byte;
begin
  Result := false;

  try
    serialSendByte(blcIdentifierCommand);
    serialSendByte(blcCommandProgramCheckCRC);
    serialSendString(AnsiChar(lo(connectionCRC)) + AnsiChar(hi(connectionCRC)));

    skipOneWireBytes(4);

    try
      reply := serialConnection.RecvByte(settingReadTimeout);
    except
      raise Exception.Create('Zeitüberschreitung');
      Exit;
    end;

  finally
    serialConnection.Purge;
    serialConnection.Flush;
  end;

  if (reply <> blcStatusSuccess) and (reply <> blcStatusFail) then
  begin
    raise Exception.Create('Abschluss der Übertragung nicht bestätigt');
    exit;
  end;

  if reply = blcStatusSuccess then
    Result := true;
end;
{$ENDREGION}

{$REGION 'Informationen auslesen'}
function TBootloaderRoutines.detectVerifySupport: boolean;
begin
  Result := detectSupport(blcCommandProgramVerify);
end;

function TBootloaderRoutines.detectCRCSupport: boolean;
begin
  Result := detectSupport(blcCommandProgramCheckCRC);
end;

function TBootloaderRoutines.readRevision: string;
var
  readBuffer: string;
  readBytes: TByteList;
begin
  Result := '';

  readBytes := readInfo(blcReadRevision);
  readBuffer := convertByteList2String(readBytes);

  if length(readBuffer) = 2 then
    readBuffer := readBuffer[1] + '.' + readBuffer[2]
  else
    raise Exception.Create('Ungültiger Datensatz empfangen (Revision)');

  Result := readBuffer;
end;

function TBootloaderRoutines.readSignature: string;
var
  readBuffer: string;
  readBytes: TByteList;
begin
  Result := '';

  readBytes := readInfo(blcReadSignature);
  readBuffer := convertByteList2HexString(readBytes, true);

  Result := readBuffer;
end;

function TBootloaderRoutines.readBufferSize: integer;
var
  readBytes: TByteList;
begin
  Result := 0;

  readBytes := readInfo(blcReadBufferSize);

  if high(readBytes) = 1 then
    Result := readBytes[1] + (readBytes[0] * 256)
  else
    raise Exception.Create('Ungültiger Datensatz empfangen (Sendpuffer-Größe)');
end;

function TBootloaderRoutines.readFlashSize: longint;
var
  readBytes: TByteList;
begin
  Result := 0;

  readBytes := readInfo(blcReadFlashSize);

  if high(readBytes) = 2 then
    Result := readBytes[2] + (readBytes[1] * 256) + (readBytes[0] * 65536)
  else
    raise Exception.Create('Ungültiger Datensatz empfangen (Flash-Größe)');
end;

function TBootloaderRoutines.getFailureAddress: string;
begin
  Result := IntToHex(failureAddress, 6);
end;

function TBootloaderRoutines.getDeviceName(signature: string): string;
var
  position: integer;
  device: string;
begin
  device := '';
  Result := defaultDeviceName;

  for position := low(deviceNames) to high(deviceNames) do
  begin
    if signature = copy(deviceNames[position], 1, 8) then
    begin
      device := copy(deviceNames[position], 10, length(deviceNames[position]));
      break;
    end;
  end;

  if device <> '' then
     Result := device;
end;
{$ENDREGION}

{$REGION 'Private Funktionen'}
{$REGION 'Bitübertragung'}
procedure TBootloaderRoutines.serialSendByte(data: byte);
begin
  serialConnection.SendByte(data);

  calculateCRC(data);
end;

procedure TBootloaderRoutines.serialSendString(data: string);
var
  position: integer;
begin
  serialConnection.SendString(data);

  for position := 1 to length(data) do
      calculateCRC(ord(data[position]));
end;

procedure TBootloaderRoutines.calculateCRC(data: byte);
var
  count: integer;
begin
  for count := 0 to 7 do
  begin
    if ((data and $01) xor (connectionCRC and $0001) <> 0) then
    begin
      connectionCRC := connectionCRC shr 1;
      connectionCRC := connectionCRC xor $A001;
    end
    else
    begin
      connectionCRC := connectionCRC shr 1;
    end;

    data := data shr 1;
  end;
end;

procedure TBootloaderRoutines.skipOneWireBytes(count: integer);
begin
  if oneWire = true then
  begin
    try
      serialConnection.RecvBufferStr(count, settingReadTimeout);
    except
    end;
  end;
end;
{$ENDREGION}

function TBootloaderRoutines.detectSupport(command: byte): boolean;
var
  reply: byte;
begin
  Result := false;

  try
    serialSendByte(blcIdentifierCommand);
    serialSendByte(command);

    skipOneWireBytes(2);

    try
      if serialConnection.RecvByte(settingReadTimeout) = blcStatusBadCommand then
      begin
        Result := false;
        exit;
      end;
    except
    end;

    serialSendByte(blcIdentifierCommand);
    serialSendByte(blcCommandProgramFinish);

    skipOneWireBytes(2);

    try
      reply := serialConnection.RecvByte(settingReadTimeout);
      if (reply = blcStatusSuccess) or (reply = blcStatusFail) then
         Result := true;
    except
    end;

  finally
    serialConnection.Purge;
    serialConnection.Flush;
  end;
end;

function TBootloaderRoutines.readInfo(readCommand: integer): TByteList;
var
  answerLength: byte;
  readAnswer: TByteList;
  readLeader: byte;
  readTrailer: byte;
  count: integer;
begin
  //Daten anfordern
  serialSendByte(blcIdentifierCommand);
  serialSendByte(readCommand);

  skipOneWireBytes(2);

  try
    try
      readLeader := serialConnection.RecvByte(settingReadTimeout);

      answerLength := serialConnection.RecvByte(settingReadTimeout) - 1;
      SetLength(readAnswer, answerLength);
      count := 0;

      while answerLength > 0 do
      begin
        readAnswer[count] := serialConnection.RecvByte(settingReadTimeout);

        inc(count);
        dec(answerLength);
      end;

      readTrailer := serialConnection.RecvByte(settingReadTimeout);
    except
      raise Exception.Create('Zeitüberschreitung');
      exit;
    end;

    if (readLeader <> blcIdentifierAnswer) or (readTrailer <> blcStatusSuccess) then
    begin
      raise Exception.Create('Ungültige Daten empfangen');
      exit;
    end;
  finally
    serialConnection.Purge;
    serialConnection.Flush;
  end;

  Result := readAnswer;
end;

function TBootloaderRoutines.uploadData(targetCommand: byte; waitForContinue: boolean; buffer: TMemoryBuffer; bufferLength: longint; chunkLength: longint; progressBar: TProgressBar): boolean;
var
  chunkCount: longint;
  chunkLeftover: longint;
  chunkData: string;
  chunkPosition: longint;
  currentChunk: longint;
  reply: byte;
begin
  chunkPosition := low(buffer);
  currentChunk := 0;
  Result := false;

  chunkCount := bufferLength div chunkLength;
  chunkLeftover := bufferLength - ((chunkCount) * chunkLength);
  if chunkLeftover <> 0 then
    inc(chunkCount);

  if Assigned(progressBar) then
  begin
    progressBar.Position := 0;
    progressBar.Max := chunkCount;
    progressBar.Step := 1;
  end;

  //Programmieren starten
  serialSendByte(blcIdentifierCommand);
  serialSendByte(targetCommand);

  skipOneWireBytes(2);

  try
    while currentChunk <= chunkCount do
    begin
      //Datenblock laden, Zählerstände setzen
      chunkData := getBufferChunk(buffer, chunkPosition, chunkLength, bufferLength);
      inc(currentChunk);
      failureAddress := chunkPosition;

      //Datenblock senden
      serialSendString(chunkData);

      //GUI-Update
      Application.ProcessMessages;

      skipOneWireBytes(Length(ChunkData));

      //Nach der Übertragung aller Chunks prüfen, ob alle Daten gesendet wurden
      //Ja: Schleife beenden, Antwort wird nicht abgewartet
      //Nein: Durch Füllzeichen stimmte die Berechnung nicht, noch einen Chunk senden
      if currentChunk = chunkCount then
      begin
        if chunkPosition < bufferLength then
          inc(chunkCount)
        else
          break;
      end;

      //Auf "Fortsetzen"-Kommando warten
      if waitForContinue = true then
      begin
        try
          reply := serialConnection.RecvByte(settingReadLongTimeout);
        except
          raise Exception.Create('Zeitüberschreitung');
          exit;
        end;

        if reply <> blcIdentifierContinue then
        begin
          raise Exception.Create('Ungültige Antwort empfangen');
          exit;
        end;
      end;

      if Assigned(progressBar) then
        progressBar.StepIt;
      Application.ProcessMessages;
    end;

    //Programmieren beenden
    serialSendByte(blcIdentifierCommand);
    serialSendByte(blcCommandProgramFinish);

    skipOneWireBytes(2);

    try
      reply := serialConnection.RecvByte(settingReadLongTimeout);
    except
      raise Exception.Create('Zeitüberschreitung');
      exit;
    end;

    if (reply <> blcStatusSuccess) and (reply <> blcStatusFail) then
    begin
      raise Exception.Create('Abschluss der Übertragung nicht bestätigt');
      exit;
    end;
  finally
    serialConnection.Purge;
    serialConnection.Flush;
  end;

  if reply = blcStatusSuccess then
    Result := true;

  if Assigned(progressBar) then
    progressBar.Position := progressBar.Max;
  Application.ProcessMessages;
end;

function TBootloaderRoutines.getBufferChunk(buffer: TMemoryBuffer; var startPosition: longint; chunkLength: longint; bufferMaxLength: longint): string;
var
  convert: string;
  convertCount: longint;
begin
  convert := '';
  convertCount := 0;
  Result := '';

  if (startPosition < 0) or (startPosition < low(buffer)) then
    startPosition := low(buffer);
  if startPosition > bufferMaxLength then
    startPosition := bufferMaxLength;
  if (chunkLength + startPosition) > bufferMaxLength then
    chunkLength := bufferMaxLength - startPosition;
  if (chunkLength <= 0) then
    Exit;

  while convertCount < chunkLength do
  begin
    if (buffer[startPosition] = blcDataIllegalChar) or (buffer[startPosition] = blcIdentifierEscape) then
    begin
      convert := convert + AnsiChar(blcIdentifierEscape);
      convert := convert + AnsiChar(buffer[startPosition] + blcDataIllegalCharShift);
    end
    else
    begin
      convert := convert + AnsiChar(buffer[startPosition]);
    end;

    inc(convertCount);
    inc(startPosition);

    if startPosition > bufferMaxLength then
       break;
  end;

  Result := convert;
end;
{$ENDREGION}

end.

